﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using Silvermoon.Controls;
using Passwords.Models;
using Silvermoon.Shapes;
using Silvermoon.Windows;
using Passwords.Controls;
using Silvermoon.Images;
using Silvermoon.UI;
using Silvermoon.Windows.Styles;
using System.Globalization;
using System.Diagnostics;
using Passwords.Biz;

namespace Passwords.Pages
{
    public class DatePickerPage : ItemPickerPage
    {
        private DatePickerMode mode = DatePickerMode.Date;
        private Picker dayPicker;
        private Picker monthPicker;
        private Picker yearPicker;
        private AnimationGrid grid;
        private DatePickerModel model;

        private DatePickerPage()
            : base()
        {
            InitializePage();
        }

        public DatePickerPage(DatePickerMode mode)
            : base()
        {
            this.mode = mode;
            InitializePage();
        }

        public DatePickerMode Mode
        {
            get { return mode; }
            private set
            {
                if (mode != value)
                {
                    mode = value;
                    ArrangeGrid();
                }
            }
        }


        public DateTime Date
        {
            get { return GetDateFromPicker(); }
            set
            {
                SetPickerDate(value);
            }
        }

        private void SetPickerDate(DateTime value)
        {
            SetDays(value.Year, value.Month, value.Day);
        }

        private DateTime GetDateFromPicker()
        {
            int year = model.GetYearFromIndex(yearPicker.SelectedItemIndex);
            int month = model.GetMonthFromIndex(monthPicker.SelectedItemIndex);
            int day = mode == DatePickerMode.Date ? model.GetDayFromIndex(dayPicker.SelectedItemIndex) : 1;
            return new DateTime(year, month, day);
        }

        protected override void Dispose(bool dispose)
        {
            base.Dispose(dispose);
        }

        private void InitializePage()
        {
            Style = StylePool.DarkEditWindow;
            SetTransition(300, TransitionMask.Fade | TransitionMask.Zoom, TransitionMask.Fade | TransitionMask.Explode);
            Title = "CHOOSE DATE";

            const PickerMode pm = PickerMode.Fade | PickerMode.AutoHide;
            Styles.SetClass<Picker>(new PickerStyle
            {
                ItemHeight = 140,
                Lense = Shape.Empty,
                //Lense = new BorderShape(3, PhoneColors.White),
                IsCyclic = true,
                Mode = pm,
                Margin = Margin.Zero,
                TextColor = PhoneColors.White
            });

            model = new DatePickerModel();

            dayPicker = new Picker();
            monthPicker = new Picker { Margin = new Margin(0, 0, 0, 0) };
            yearPicker = new Picker();

            grid = new AnimationGrid { Margin = Margin.Zero };

            grid.AddControl(model.GetDayIndex(), dayPicker);
            grid.AddControl(model.GetMonthIndex(), monthPicker);
            grid.AddControl(model.GetYearIndex(), yearPicker);
            Controls.Add(grid);

            BindModel();
            Date = DateTime.Now.Date;
            ArrangeGrid();
        }


        protected override void OnShown()
        {
            monthPicker.SelectedItemChanged += new EventHandler(OnSelectedItemChanged);
            yearPicker.SelectedItemChanged += new EventHandler(OnSelectedItemChanged);
            dayPicker.SelectedItemChanged += (s, e) => IsSaveEnabled = true;
            base.OnShown();
        }

        private void ArrangeGrid()
        {
            switch (Mode)
            {
                case DatePickerMode.Date: ArrangeDate(); break;
                case DatePickerMode.MonthYear: ArrangeMonthYear(); break;
                default: throw new NotSupportedException();
            }
        }

        private void ArrangeDate()
        {
            GridDefinition def = new GridDefinition();
            def.Columns.AddRange(2f, 2f, 2f);
            def.Rows.AddRange(1f);
            def.CreateAutoCells();

            grid.Controls.Clear();
            grid.Definition = def;
            grid.AddControl(model.GetDayIndex(), dayPicker);
            grid.AddControl(model.GetMonthIndex(), monthPicker);
            grid.AddControl(model.GetYearIndex(), yearPicker);
        }

        private void ArrangeMonthYear()
        {
            GridDefinition def = new GridDefinition();
            def.Columns.AddRange(1f, 2f, 2f, 1f);
            def.Rows.AddRange(1f);
            def.CreateAutoCells();

            grid.Controls.Clear();
            grid.Definition = def;
            grid.AddControl(1, monthPicker);
            grid.AddControl(2, yearPicker);
        }

        void OnSelectedItemChanged(object sender, EventArgs e)
        {
            if (yearPicker.SelectedItemIndex >= 0 && monthPicker.SelectedItemIndex >= 0 && dayPicker.SelectedItemIndex >= 0)
            {
                int year = model.GetYearFromIndex(yearPicker.SelectedItemIndex);
                int month = model.GetMonthFromIndex(monthPicker.SelectedItemIndex);
                int day = model.GetDayFromIndex(dayPicker.SelectedItemIndex);
                SetDays(year, month, day);
            }
            IsSaveEnabled = true;
        }

        private void BindModel()
        {

            dayPicker.GetVirtualItem += OnGetPickerItem;
            monthPicker.GetVirtualItem += OnGetPickerItem;
            yearPicker.GetVirtualItem += OnGetPickerItem;

            monthPicker.ItemsSource = model.Months;
            yearPicker.ItemsSource = model.Years;

        }

        private void SetDays(int year, int month, int day)
        {
            monthPicker.SelectedItemIndex = model.GetIndexFromMonth(month);
            yearPicker.SelectedItemIndex = model.GetIndexFromYear(year);
            if (mode == DatePickerMode.Date)
            {
                dayPicker.ItemsSource = model.GetDays(month, year);
                dayPicker.SelectedItemIndex = model.GetIndexFromDay(day);
            }
        }

        void OnGetPickerItem(object sender, PickerItemEventArgs e)
        {
            var item = DatePickerItem.GetItem();
            item.DataValue = e.Value;
            e.Container = item;
        }

        protected override void OnPasswordItemChanged(PasswordItem item)
        {
            if (item != null)
            {
                switch (mode)
                {
                    case DatePickerMode.Date: Date = item.Value.StringToDate(); break;
                    case DatePickerMode.MonthYear: Date = item.Value.StringToMonthYear(); break;
                }
                if (item.IsNew || string.IsNullOrEmpty(item.Value)) IsSaveEnabled = true;
            }
            else
            {
                Date = DateTime.Now;
            }
            base.OnPasswordItemChanged(item);
        }

        protected override void OnOK(object sender, CommandEventArgs e)
        {
            if (PasswordItem != null)
            {
                switch (mode)
                {
                    case DatePickerMode.Date:
                        PasswordItem.Value = Date.DateToString();
                        break;

                    case DatePickerMode.MonthYear:
                        PasswordItem.Value = Date.MonthYearToString();
                        break;
                }
            }
            base.OnOK(sender, e);
        }
    }
}
