﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using Silvermoon.Controls;
using Silvermoon.Shapes;
using Silvermoon.Images;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Silvermoon.UI;
using Silvermoon.Controls.Toolbars;
using Silvermoon.Windows.Styles;
using System.Drawing;
using System.Collections;

namespace Silvermoon.Debug
{
    public class DebuggerWindow : Window
    {

        private static List<Control> allocatedControls = new List<Control>();
        private Label memoryLabel;
        private Label availablePhysicalLabel;
        private Label controlCountLabel;
        private Label shapeCountLabel;
        private Label allocedControlLabel;
        static GridDefinition buttonGridDef;
        private static bool monitor;

        /// <summary>
        /// Gets a list of all currently monitored controls.
        /// See also <seealso cref="Monitor"/>.
        /// </summary>
        public static List<Control> AllocatedControls
        {
            get { return allocatedControls; }
        }

        public static void ClearAllocatedControlList()
        {
            allocatedControls.Clear();
        }

        /// <summary>
        /// If set to true, the debugger monitors the creation and disposal of all controls and holds them in a list.
        /// See also <seealso cref="MonitoredControls"/>
        /// </summary>
        public static bool Monitor
        {
            get { return monitor; }
            set
            {
                if (monitor != value)
                {
                    monitor = value;
                    if (value)
                    {
                        Control.ControlCreated += new EventHandler(Control_ControlCreated);
                        Control.ControlDisposed += new EventHandler(Control_ControlDisposed);
                    }
                    else
                    {
                        Control.ControlCreated -= new EventHandler(Control_ControlCreated);
                        Control.ControlDisposed -= new EventHandler(Control_ControlDisposed);
                    }
                }
            }
        }

        static DebuggerWindow()
        {

            buttonGridDef = new GridDefinition();
            buttonGridDef.Rows.AddRange(80, 20, 80, 20, 80, 20, 80, 20, 80, 20);
            buttonGridDef.Columns.AddRange(1f);
            for (int i = 0; i < 10; i++) buttonGridDef.Cells.Add(new GridCell((i / 2) * 2, 0, 1, 1));

        }

        static void Control_ControlDisposed(object sender, EventArgs e)
        {
            allocatedControls.Remove(sender as Control);
        }

        static void Control_ControlCreated(object sender, EventArgs e)
        {
            allocatedControls.Add(sender as Control);
        }

        public static void ShowDebugInformation()
        {
            bool monitor = Monitor;
            DebuggerWindow.Monitor = false;
            DebuggerWindow info = new DebuggerWindow(monitor);
            info.Show();
        }

        private bool monitorAfter;

        public DebuggerWindow(bool monitor)
            : base()
        {
            monitorAfter = monitor;
            Initialize();
            Refresh();
        }

        protected override void Dispose(bool dispose)
        {
            Monitor = monitorAfter ;
            base.Dispose(dispose);
        }

        private void Initialize()
        {
            SetTransition(300, TransitionMask.Fade, TransitionMask.Fade);
            Monitor = false;
            Title = "DEBUG INFORMATION";
            AnimationGrid grid = new AnimationGrid { Margin = new Margin(20, 16, 20, 0) };
            grid.Definition = buttonGridDef;

            grid.AddControl(0, new Label { Style = StylePool.SmallDebugLabel, Text = "Allocated .NET Memory" });
            grid.AddControl(2, new Label { Style = StylePool.SmallDebugLabel, Text = "Available Physical Memory" });
            grid.AddControl(4, new Label { Style = StylePool.SmallDebugLabel, Text = "# Total Controls used" });
            grid.AddControl(6, new Label { Style = StylePool.SmallDebugLabel, Text = "# Total Shapes used" });
            grid.AddControl(8, new Label { Style = StylePool.SmallDebugLabel, Text = "# Controls monitored" });
            memoryLabel = new Label { Style = StylePool.LargeDebugLabel, };
            controlCountLabel = new Label { Style = StylePool.LargeDebugLabel };
            shapeCountLabel = new Label { Style = StylePool.LargeDebugLabel };
            allocedControlLabel = new Label { Style = StylePool.LargeDebugLabel };
            availablePhysicalLabel = new Label { Style = StylePool.LargeDebugLabel };

            grid.AddControl(1, memoryLabel);
            grid.AddControl(3, availablePhysicalLabel);
            grid.AddControl(5, controlCountLabel);
            grid.AddControl(7, shapeCountLabel);
            grid.AddControl(9, allocedControlLabel);

            Controls.Add(grid);

            Toolbar tb = new Toolbar();
            ToolbarButton tbn;

            tbn = new ToolbarButton { Content = SysImages.Back };
            tbn.Tap += (s, e) => Hide();
            tb.Buttons.Add(tbn);

            tbn = new ToolbarButton { Content = SysImages.RecycleBin };
            tbn.Tap += new EventHandler<TapEventArgs>(OnClearList);
            tb.Buttons.Add(tbn);

            tbn = new ToolbarButton { Content = "?" };
            tbn.Tap += new EventHandler<TapEventArgs>(OnShowList);
            tb.Buttons.Add(tbn);


            Toolbar = tb;
        }



        void OnShowList(object sender, TapEventArgs e)
        {
            ShowYielded();
        }

        void OnClearList(object sender, TapEventArgs e)
        {
            Reset();
            GC.Collect();
            GC.WaitForPendingFinalizers();
        }

        public void Refresh()
        {
            GC.Collect();
            GC.WaitForPendingFinalizers();
            long total = GC.GetTotalMemory(true);

            NativeMemory.MemoryStatus status = NativeMemory.GetMemoryStatus();

            availablePhysicalLabel.Text = NativeMemory.FormatSize(status.AvailablePhysical);

            memoryLabel.Text = NativeMemory.FormatSize((int)total);
            controlCountLabel.Text = Control.UsedControls.ToString();
            shapeCountLabel.Text = Shape.UsedShapes.ToString();
            allocedControlLabel.Text = allocatedControls.Count.ToString();
        }



        private void Reset()
        {
            allocatedControls.Clear();
            Refresh();
        }

        private void Enumerate()
        {
            Refresh();
            foreach (Control c in allocatedControls)
            {
                string type = c.GetType().Name;
                int index = c.Identity;
                Trace.WriteLine(type + ": " + index.ToString());
            }
        }

        private class ControlInfo
        {
            public Control Control;
            public ControlInfo(Control c)
            {
                Control = c;
            }
        }

        private void ShowYielded()
        {
            Window w = new Window { Title = "DEBUG INFORMATION" };
            ListView listView = new ListView { Margin = new Margin(0, 24, 0, 0) };
            w.Controls.Add(listView);
            var controller = new ListViewController<TitleDetailsImageItem, ControlInfo>(listView);
            Toolbar tb = new Toolbar();
            w.Toolbar = tb;

            ToolbarButton btn = new ToolbarButton { Content = SysImages.Back };
            btn.Tap += (s, e) => w.Hide();
            tb.Buttons.Add(btn);

            controller.UpdateItem += (s, e) =>
                {
                    TitleDetailsImageItem item = e.Item;
                    ControlInfo ci = e.Value as ControlInfo;
                    Control c = ci.Control;
                    StringBuilder sb = new StringBuilder();
                    Control parent = c.Parent;
                    if (parent != null) sb.Append(parent.GetType().Name + ": " + parent.Identity.ToString());
                    if (c != null)
                    {
                        item.Title = c.GetType().Name + ": " + c.Identity.ToString();
                        item.Height = 120;
                        item.Image = SysImages.RadioBackground;
                        item.ImageSize = new Size(24, 24);
                        Label l = c as Label;
                        string details = "";
                        if (l != null) details = l.Text;
                        else
                        {
                            Button b = c as Button;
                            if (b != null)
                            {
                                StringShape ss = b.Content as StringShape;
                                if (ss != null) details = ss.Text;
                            }
                        }
                        if (!string.IsNullOrEmpty(details))
                        {
                            if (sb.Length > 0) sb.AppendLine();
                            sb.Append(details);
                        }
                    }
                    else item.Title = "{unknown}";
                    if (sb.Length == 0) sb.Append("no details");
                    item.Details = sb.ToString(); ;
                };

            List<ControlInfo> list = new List<ControlInfo>();
            foreach (var c in allocatedControls)
            {
                if (!c.IsDisposed) list.Add(new ControlInfo(c));
                //list.Add(new ControlInfo(c));
            }
            w.Shown += (s, e) => controller.DataSource = list;
            w.Back += (s, e) => w.Hide();
            w.Hidden += (s, e) => w.Dispose();

            w.Show();
        }

        protected override void OnHidden()
        {
            base.OnHidden();
            Dispose();
        }

        protected override void OnBack(HandledEventArgs e)
        {
            Hide();
        }

    }
}
